import logging
from typing import Any, Union

import discord
import discord.member

from modules.database import col_users, col_warnings
from modules.utils import config_get

logger = logging.getLogger(__name__)


class NotEnoughMoneyError(Exception):
    """User does not have enough money to do that"""

    pass


class UserNotFoundError(Exception):
    """HoloUser could not find user with such an ID in database"""

    def __init__(self, user, user_id):
        self.user = user
        self.user_id = user_id
        super().__init__(
            f"User of type {type(self.user)} with id {self.user_id} was not found"
        )


class HoloUser:
    def __init__(
        self, user: Union[discord.User, discord.Member, discord.member.Member, int]
    ) -> None:
        """Get an object that has a proper binding between Discord ID and database

        ### Args:
            * `user` (Union[discord.User, discord.Member, discord.member.Member, int]): Object from which ID can be extracted

        ### Raises:
            * `UserNotFoundError`: User with such ID does not seem to exist in database
        """

        if hasattr(user, "id"):
            self.id = user.id  # type: ignore
        else:
            self.id = user

        jav_user = col_users.find_one({"user": self.id})

        if jav_user is None:
            raise UserNotFoundError(user=user, user_id=self.id)

        self.db_id = jav_user["_id"]

        self.customrole = jav_user["customrole"]
        self.customchannel = jav_user["customchannel"]
        self.warnings = self.warns()

    def warns(self) -> int:
        """Get number of warnings user has

        ### Returns:
            * `int`: Number of warnings
        """
        warns = col_warnings.find_one({"user": self.id})
        if warns == None:
            return 0
        else:
            return warns["warns"]

    def warn(self, count=1, reason: str = "Not provided") -> None:
        """Warn and add count to warns number

        ### Args:
            * `count` (int, optional): Count of warnings to be added. Defaults to 1.
        """
        warns = col_warnings.find_one({"user": self.id})
        if warns != None:
            col_warnings.update_one(
                filter={"_id": self.db_id},
                update={"$set": {"warns": warns["warns"] + count}},
            )
        else:
            col_warnings.insert_one(document={"user": self.id, "warns": count})
        logger.info(f"User {self.id} was warned {count} times due to: {reason}")

    def set(self, key: str, value: Any) -> None:
        """Set attribute data and save it into database

        ### Args:
            * `key` (str): Attribute to be changed
            * `value` (Any): Value to set
        """
        if not hasattr(self, key):
            raise AttributeError()
        setattr(self, key, value)
        col_users.update_one(
            filter={"_id": self.db_id}, update={"$set": {key: value}}, upsert=True
        )
        logger.info(f"Set attribute {key} of user {self.id} to {value}")

    async def is_moderator(
        self, member: Union[discord.User, discord.Member, discord.member.Member]
    ) -> bool:
        """Check if user is moderator or council member

        ### Args:
            * `member` (Union[discord.User, discord.Member, discord.member.Member]): Member object

        ### Returns:
            `bool`: `True` if member is a moderator or member of council and `False` if not
        """
        if isinstance(member, discord.User):
            return False
        moderator_role = await config_get("moderators", "roles")
        council_role = await config_get("council", "roles")
        for role in member.roles:
            if role.id == moderator_role or role.id == council_role:
                return True
        return False

    async def is_council(
        self, member: Union[discord.User, discord.Member, discord.member.Member]
    ) -> bool:
        """Check if user is a member of council

        ### Args:
            * `member` (Union[discord.User, discord.Member, discord.member.Member]): Member object

        ### Returns:
            `bool`: `True` if member is a member of council and `False` if not
        """
        if isinstance(member, discord.User):
            return False
        council_role = await config_get("council", "roles")
        for role in member.roles:
            if role.id == council_role:
                return True
        return False

    # def purge(self) -> None:
    #     """Completely remove data from database. Will not remove transactions logs and warnings."""
    #     col_users.delete_one(filter={"_id": self.db_id})
    #     self.unauthorize()
